use std::{error::Error, fmt::Display, rc::Rc};

use crate::func::*;

pub trait Context {
    type T: Monad;

    type LookupError: Error;

    fn hash(s: &[u8]) -> Hash;
}

pub enum ResolutionError<L, P> {
    Lookup(L),
    Parse(P),
}

pub type ResolutionResult<'a, Ctx, A> = Result<
    Rc<A>,
    ResolutionError<
        <Ctx as Context>::LookupError,
        <<A as Mentionable<'a, Ctx>>::Fctr as Factory<'a, Ctx>>::ParseError,
    >,
>;

pub type Wrapped<'a, Ctx, A> = <<Ctx as Context>::T as WeakFunctor>::F<'a, A>;

pub type Resolution<'a, Ctx, A> = Wrapped<'a, Ctx, ResolutionResult<'a, Ctx, A>>;

pub const HASH_SIZE: usize = 32;

pub const HASH_ZEROS: [u8; HASH_SIZE] = [0; HASH_SIZE];

pub type Hash = [u8; HASH_SIZE];

pub trait Serializer {
    fn write(&mut self, buf: &[u8]);
    fn tell(&self) -> usize;
}

impl Serializer for Vec<u8> {
    fn write(&mut self, buf: &[u8]) {
        self.extend(buf);
    }

    fn tell(&self) -> usize {
        self.len()
    }
}

pub trait Serializable {
    fn serialize(&self, serializer: &mut dyn Serializer);
}

pub trait Mentionable<'a, Ctx: 'a + Context>: 'a + Serializable {
    type Fctr: Factory<'a, Ctx, Mtbl = Self>;

    fn factory(&self) -> Self::Fctr;
    fn topology(&self) -> Hash {
        let mut vec = Vec::new();
        for point in self.points() {
            vec.extend(point.point);
        }
        Ctx::hash(&vec)
    }
    fn points(&self) -> Vec<Point<'a, Ctx, TypelessMentionable<'a, Ctx>>>;
}

pub type HashResolution<'a, Ctx> =
    Wrapped<'a, Ctx, Result<(Vec<u8>, Rc<dyn Resolver<'a, Ctx>>), <Ctx as Context>::LookupError>>;

#[derive(Clone, Copy)]
pub struct Address {
    pub point: Hash,
    pub index: usize,
}

pub trait Resolver<'a, Ctx: 'a + Context>: 'a {
    fn resolve(self: Rc<Self>, address: Address) -> HashResolution<'a, Ctx>;
}

pub trait Deserializer {
    fn read_n(&mut self, n: usize) -> &[u8];
    fn read_all(&mut self) -> &[u8];
    fn tell(&self) -> usize;
}

pub type ParseResult<'a, Ctx, F> =
    Result<<F as Factory<'a, Ctx>>::Mtbl, <F as Factory<'a, Ctx>>::ParseError>;

pub trait Factory<'a, Ctx: 'a + Context>: Clone + 'a {
    type Mtbl: Mentionable<'a, Ctx, Fctr = Self>;
    type ParseError: 'a + Error;

    fn deserialize(
        &self,
        deserializer: &mut dyn Deserializer,
        resolver: Rc<dyn Resolver<'a, Ctx>>,
    ) -> ParseResult<'a, Ctx, Self>;

    fn unexpected_tail(&self, tail: &[u8]) -> Self::ParseError;
}

pub trait Origin<'a, Ctx: 'a + Context>: 'a {
    type Mtbl: Mentionable<'a, Ctx>;
    fn factory(&self) -> <Self::Mtbl as Mentionable<'a, Ctx>>::Fctr;
    fn resolve(self: Rc<Self>) -> Resolution<'a, Ctx, Self::Mtbl>;
}

struct LocalOrigin<A>(Rc<A>);

impl<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> Origin<'a, Ctx> for LocalOrigin<A> {
    type Mtbl = A;

    fn factory(&self) -> <Self::Mtbl as Mentionable<'a, Ctx>>::Fctr {
        self.0.factory()
    }

    fn resolve(self: Rc<Self>) -> Resolution<'a, Ctx, Self::Mtbl> {
        <Ctx::T as Applicative>::pure(Ok(self.0.clone()))
    }
}

impl<A> From<A> for LocalOrigin<A> {
    fn from(value: A) -> Self {
        LocalOrigin(value.into())
    }
}

pub struct Point<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> {
    pub point: Hash,
    pub origin: Rc<dyn Origin<'a, Ctx, Mtbl = A>>,
}

impl<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> Clone for Point<'a, Ctx, A> {
    fn clone(&self) -> Self {
        Self {
            point: self.point,
            origin: self.origin.clone(),
        }
    }
}

impl<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> Point<'a, Ctx, A> {
    fn prepare_bytes_for_hashing(mentioned: &A) -> Vec<u8> {
        let mut vec = mentioned.topology().to_vec();
        mentioned.serialize(&mut vec);
        vec
    }

    fn from_fields(point: Hash, origin: Rc<dyn Origin<'a, Ctx, Mtbl = A>>) -> Self {
        Point { point, origin }
    }

    fn from_values<O: Origin<'a, Ctx, Mtbl = A>>(point: Hash, origin: O) -> Self {
        Self::from_fields(point, Rc::new(origin))
    }

    pub fn resolve(&self) -> Resolution<'a, Ctx, A> {
        self.origin.clone().resolve()
    }
}

impl<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> From<A> for Point<'a, Ctx, A> {
    fn from(value: A) -> Self {
        Self::from_values(
            Ctx::hash(Self::prepare_bytes_for_hashing(&value).as_slice()),
            LocalOrigin::from(value),
        )
    }
}

type TypelessSerialize<'a> = dyn 'a + Fn(&mut dyn Serializer);

pub struct TypelessMentionable<'a, Ctx: 'a + Context> {
    t_serialize: Box<TypelessSerialize<'a>>,
    t_factory: TypelessFactory<'a, Ctx>,
    t_topology: Hash,
    t_points: Vec<Point<'a, Ctx, TypelessMentionable<'a, Ctx>>>,
}

type TypelessParsed<'a, Ctx> = Result<TypelessMentionable<'a, Ctx>, Box<dyn 'a + Error>>;

type TypelessDeserialize<'a, Ctx> =
    dyn 'a + Fn(&mut dyn Deserializer, Rc<dyn Resolver<'a, Ctx>>) -> TypelessParsed<'a, Ctx>;

type TypelessUnexpectedTail<'a> = dyn 'a + Fn(&[u8]) -> TypelessError<'a>;

pub struct TypelessFactory<'a, Ctx: 'a + Context> {
    t_deserialize: Rc<TypelessDeserialize<'a, Ctx>>,
    t_unexpected_tail: Rc<TypelessUnexpectedTail<'a>>,
}

impl<'a, Ctx: 'a + Context> Serializable for TypelessMentionable<'a, Ctx> {
    fn serialize(&self, serializer: &mut dyn Serializer) {
        (self.t_serialize)(serializer);
    }
}

impl<'a, Ctx: 'a + Context> Mentionable<'a, Ctx> for TypelessMentionable<'a, Ctx> {
    type Fctr = TypelessFactory<'a, Ctx>;

    fn factory(&self) -> Self::Fctr {
        self.t_factory.clone()
    }

    fn topology(&self) -> Hash {
        self.t_topology
    }

    fn points(&self) -> Vec<Point<'a, Ctx, TypelessMentionable<'a, Ctx>>> {
        self.t_points.clone()
    }
}

impl<'a, Ctx: 'a + Context> Clone for TypelessFactory<'a, Ctx> {
    fn clone(&self) -> Self {
        Self {
            t_deserialize: self.t_deserialize.clone(),
            t_unexpected_tail: self.t_unexpected_tail.clone(),
        }
    }
}

#[derive(Debug)]
pub struct TypelessError<'a>(Box<dyn 'a + Error>);

impl<'a> Display for TypelessError<'a> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.0.fmt(f)
    }
}

impl<'a> Error for TypelessError<'a> {}

impl<'a, Ctx: 'a + Context> Factory<'a, Ctx> for TypelessFactory<'a, Ctx> {
    type Mtbl = TypelessMentionable<'a, Ctx>;

    type ParseError = TypelessError<'a>;

    fn deserialize(
        &self,
        deserializer: &mut dyn Deserializer,
        resolver: Rc<dyn Resolver<'a, Ctx>>,
    ) -> ParseResult<'a, Ctx, Self> {
        match (self.t_deserialize)(deserializer, resolver) {
            Ok(mentionable) => Ok(mentionable),
            Err(error) => Err(TypelessError(error)),
        }
    }

    fn unexpected_tail(&self, tail: &[u8]) -> Self::ParseError {
        (self.t_unexpected_tail)(tail)
    }
}

impl<'a, Ctx: 'a + Context> TypelessMentionable<'a, Ctx> {
    pub fn from_typed<A: Mentionable<'a, Ctx>>(mentionable: Rc<A>) -> Self {
        let factory = TypelessFactory::from_typed(mentionable.factory());
        let topology = mentionable.topology();
        let points = mentionable.points();
        TypelessMentionable {
            t_serialize: Box::new(move |serializer| mentionable.serialize(serializer)),
            t_factory: factory,
            t_topology: topology,
            t_points: points,
        }
    }
}

impl<'a, Ctx: 'a + Context> TypelessFactory<'a, Ctx> {
    pub fn from_typed<F: Factory<'a, Ctx>>(factory: F) -> Self {
        let tail_factory = factory.clone();
        TypelessFactory {
            t_deserialize: Rc::new(move |deserializer, resolver| {
                match factory.deserialize(deserializer, resolver) {
                    Ok(mentionable) => Ok(TypelessMentionable::from_typed(Rc::new(mentionable))),
                    Err(error) => Err(Box::new(error)),
                }
            }),
            t_unexpected_tail: Rc::new(move |tail| {
                TypelessError::from_typed(tail_factory.unexpected_tail(tail))
            }),
        }
    }
}

impl<'a> TypelessError<'a> {
    pub fn from_typed<E: 'a + Error>(error: E) -> Self {
        TypelessError(Box::new(error))
    }
}
