pub mod atomic;
pub mod cast;
pub mod inlining;
pub mod nullable;

use std::cmp::min;
use std::{error::Error, fmt::Display, rc::Rc};

use crate::core::*;
use crate::func::*;

struct WrappedOrigin<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> {
    w_factory: A::Fctr,
    w_resolve: Box<dyn 'a + Fn() -> Resolution<'a, Ctx, A>>,
}

trait MappableOrigin<'a, Ctx: 'a + Context>: Origin<'a, Ctx> {
    fn map<B: Mentionable<'a, Ctx>>(
        self: Rc<Self>,
        map_ok: impl 'a + Clone + Fn(Rc<Self::Mtbl>) -> B,
        map_err: impl 'a
            + Clone
            + Fn(
                <<Self::Mtbl as Mentionable<'a, Ctx>>::Fctr as Factory<'a, Ctx>>::ParseError,
            ) -> <B::Fctr as Factory<'a, Ctx>>::ParseError,
        map_factory: impl 'a + FnOnce(<Self::Mtbl as Mentionable<'a, Ctx>>::Fctr) -> B::Fctr,
    ) -> WrappedOrigin<'a, Ctx, B>;
}

fn map_resolve<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>, B: Mentionable<'a, Ctx>>(
    resolve: impl 'a + Fn() -> Resolution<'a, Ctx, A>,
    map_ok: impl 'a + Fn(Rc<A>) -> B,
    map_err: impl 'a
        + Fn(<A::Fctr as Factory<'a, Ctx>>::ParseError) -> <B::Fctr as Factory<'a, Ctx>>::ParseError,
) -> Resolution<'a, Ctx, B> {
    Ctx::T::fmap(
        move |resolved| match resolved {
            Ok(mentionable) => Ok(Rc::new(map_ok(mentionable))),
            Err(ResolutionError::Parse(parse_error)) => {
                Err(ResolutionError::Parse(map_err(parse_error)))
            }
            Err(ResolutionError::Lookup(lookup_error)) => {
                Err(ResolutionError::Lookup(lookup_error))
            }
        },
        resolve(),
    )
}

impl<'a, Ctx: 'a + Context, O: ?Sized + Origin<'a, Ctx>> MappableOrigin<'a, Ctx> for O {
    fn map<B: Mentionable<'a, Ctx>>(
        self: Rc<Self>,
        map_ok: impl 'a + Clone + Fn(Rc<Self::Mtbl>) -> B,
        map_err: impl 'a
            + Clone
            + Fn(
                <<Self::Mtbl as Mentionable<'a, Ctx>>::Fctr as Factory<'a, Ctx>>::ParseError,
            ) -> <B::Fctr as Factory<'a, Ctx>>::ParseError,
        map_factory: impl 'a + FnOnce(<Self::Mtbl as Mentionable<'a, Ctx>>::Fctr) -> B::Fctr,
    ) -> WrappedOrigin<'a, Ctx, B> {
        let origin = self.clone();
        WrappedOrigin {
            w_factory: map_factory(self.factory()),
            w_resolve: Box::new(move || {
                let origin = origin.clone();
                let map_ok = map_ok.clone();
                let map_err = map_err.clone();
                map_resolve(move || origin.clone().resolve(), map_ok, map_err)
            }),
        }
    }
}

impl<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> Origin<'a, Ctx> for WrappedOrigin<'a, Ctx, A> {
    type Mtbl = A;

    fn factory(&self) -> <Self::Mtbl as Mentionable<'a, Ctx>>::Fctr {
        self.w_factory.clone()
    }

    fn resolve(self: Rc<Self>) -> Resolution<'a, Ctx, A> {
        (self.w_resolve)()
    }
}

impl<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> Point<'a, Ctx, A> {
    pub fn typeless(&self) -> Point<'a, Ctx, TypelessMentionable<'a, Ctx>> {
        Point {
            point: self.point,
            origin: Rc::new(self.origin.clone().map(
                TypelessMentionable::from_typed,
                TypelessError::from_typed,
                TypelessFactory::from_typed,
            )),
        }
    }
}

impl<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> Serializable for Point<'a, Ctx, A> {
    fn serialize(&self, serializer: &mut dyn Serializer) {
        serializer.write(&self.point)
    }
}

impl<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> Mentionable<'a, Ctx> for Point<'a, Ctx, A> {
    type Fctr = PointFactory<A::Fctr>;

    fn factory(&self) -> Self::Fctr {
        PointFactory {
            factory: self.origin.factory(),
        }
    }

    fn topology(&self) -> Hash {
        Ctx::hash(&self.point)
    }

    fn points(&self) -> Vec<Point<'a, Ctx, TypelessMentionable<'a, Ctx>>> {
        vec![self.typeless()]
    }
}

#[derive(Clone)]
pub struct PointFactory<F> {
    factory: F,
}

#[derive(Debug)]
pub enum PointParseError {
    WrongLength(usize),
}

impl Display for PointParseError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            PointParseError::WrongLength(length) => f.write_fmt(format_args!(
                "expected {} bytes, received {}",
                HASH_SIZE, length
            )),
        }
    }
}

impl Error for PointParseError {}

impl From<&[u8]> for PointParseError {
    fn from(value: &[u8]) -> Self {
        PointParseError::WrongLength(value.len())
    }
}

impl<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> Point<'a, Ctx, A> {
    fn from_point(address: Address, factory: A::Fctr, resolver: Rc<dyn Resolver<'a, Ctx>>) -> Self {
        Point {
            point: address.point,
            origin: Rc::new(ResolverOrigin {
                r_factory: factory,
                r_address: address,
                r_resolver: resolver,
            }),
        }
    }
}

impl<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> AsRef<[u8]> for Point<'a, Ctx, A> {
    fn as_ref(&self) -> &[u8] {
        &self.point
    }
}

pub struct Addresses {
    current: usize,
}

impl Addresses {
    fn next<'a>(&mut self, deserializer: &'a mut dyn Deserializer) -> Result<Address, &'a [u8]> {
        let point = deserializer.read_n_const::<HASH_SIZE>()?;
        let address = Address {
            point,
            index: self.current,
        };
        self.current += 1;
        Ok(address)
    }

    fn next_point<'a, 'b, Ctx: 'a + Context, A: Mentionable<'a, Ctx>>(
        &mut self,
        deserializer: &'b mut dyn Deserializer,
        resolver: Rc<dyn Resolver<'a, Ctx>>,
        factory: A::Fctr,
    ) -> Result<Point<'a, Ctx, A>, &'b [u8]> {
        Ok(Point::from_point(
            self.next(deserializer)?,
            factory,
            resolver,
        ))
    }

    fn start() -> Self {
        Addresses { current: 0 }
    }
}

impl<'a, Ctx: 'a + Context, F: Factory<'a, Ctx>> Factory<'a, Ctx> for PointFactory<F> {
    type Mtbl = Point<'a, Ctx, F::Mtbl>;

    type ParseError = PointParseError;

    fn deserialize(
        &self,
        deserializer: &mut dyn Deserializer,
        resolver: Rc<dyn Resolver<'a, Ctx>>,
    ) -> ParseResult<'a, Ctx, Self> {
        let mut addresses = Addresses::start();
        Ok(addresses.next_point(deserializer, resolver, self.factory.clone())?)
    }

    fn unexpected_tail(&self, tail: &[u8]) -> Self::ParseError {
        PointParseError::WrongLength(HASH_SIZE + tail.len())
    }
}

struct ResolverOrigin<'a, Ctx: 'a + Context, F: Factory<'a, Ctx>> {
    r_factory: F,
    r_address: Address,
    r_resolver: Rc<dyn Resolver<'a, Ctx>>,
}

fn _resolve_origin<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>>(
    origin: Rc<ResolverOrigin<'a, Ctx, A::Fctr>>,
) -> Resolution<'a, Ctx, A> {
    let resolution = origin.r_resolver.clone().resolve(origin.r_address);
    Ctx::T::fmap(
        move |resolved| match resolved {
            Ok((src, resolver)) => match origin.r_factory.parse_slice(&src, resolver) {
                Ok(mentionable) => Ok(Rc::new(mentionable)),
                Err(parse_error) => Err(ResolutionError::Parse(parse_error)),
            },
            Err(lookup_error) => Err(ResolutionError::Lookup(lookup_error)),
        },
        resolution,
    )
}

impl<'a, Ctx: 'a + Context, F: Factory<'a, Ctx>> Origin<'a, Ctx> for ResolverOrigin<'a, Ctx, F> {
    type Mtbl = F::Mtbl;

    fn factory(&self) -> <Self::Mtbl as Mentionable<'a, Ctx>>::Fctr {
        self.r_factory.clone()
    }

    fn resolve(self: Rc<Self>) -> Resolution<'a, Ctx, Self::Mtbl> {
        _resolve_origin(self)
    }
}

struct SliceDeserializer<'a> {
    slice: &'a [u8],
    pos: usize,
}

impl<'a> From<&'a [u8]> for SliceDeserializer<'a> {
    fn from(value: &'a [u8]) -> Self {
        SliceDeserializer {
            slice: value,
            pos: 0,
        }
    }
}

impl<'a> Deserializer for SliceDeserializer<'a> {
    fn read_n(&mut self, n: usize) -> &[u8] {
        let (left, right) = self.slice.split_at(min(n, self.slice.len()));
        self.slice = right;
        self.pos += left.len();
        left
    }

    fn read_all(&mut self) -> &[u8] {
        let left = self.slice;
        self.slice = &[];
        self.pos += left.len();
        left
    }

    fn tell(&self) -> usize {
        self.pos
    }
}

fn _parse_slice<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>>(
    factory: &A::Fctr,
    slice: &[u8],
    resolver: Rc<dyn Resolver<'a, Ctx>>,
) -> ParseResult<'a, Ctx, A::Fctr> {
    let mut deserializer = SliceDeserializer::from(slice);
    let mentionable = factory.deserialize(&mut deserializer, resolver)?;
    let tail = deserializer.read_all();
    if tail.is_empty() {
        Ok(mentionable)
    } else {
        Err(factory.unexpected_tail(tail))
    }
}

pub trait ExtFactory<'a, Ctx: 'a + Context>: Factory<'a, Ctx> {
    fn parse_slice(
        &self,
        slice: &[u8],
        resolver: Rc<dyn Resolver<'a, Ctx>>,
    ) -> ParseResult<'a, Ctx, Self>;
}

impl<'a, Ctx: 'a + Context, F: Factory<'a, Ctx>> ExtFactory<'a, Ctx> for F {
    fn parse_slice(
        &self,
        slice: &[u8],
        resolver: Rc<dyn Resolver<'a, Ctx>>,
    ) -> ParseResult<'a, Ctx, Self> {
        _parse_slice::<Ctx, F::Mtbl>(self, slice, resolver)
    }
}

pub trait ExtDeserializer {
    fn read_n_const<const N: usize>(&mut self) -> Result<[u8; N], &[u8]>;
}

impl<D: ?Sized + Deserializer> ExtDeserializer for D {
    fn read_n_const<const N: usize>(&mut self) -> Result<[u8; N], &[u8]> {
        let slice = self.read_n(N);
        match slice.try_into() {
            Ok(array) => Ok(array),
            Err(_) => Err(slice),
        }
    }
}
