//! This module introduces [Option]-like concepts into RADN typesystem using [Nullable].

use crate::core::*;
use crate::std::*;

/// Nullable reference type. Made for use as a linking element in data structures.
pub enum Nullable<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> {
    Null(A::Fctr),
    NotNull(Point<'a, Ctx, A>),
}

/// Nullable reference factory.
pub struct NullableFactory<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> {
    factory: A::Fctr,
}

impl<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> Serializable for Nullable<'a, Ctx, A> {
    fn serialize(&self, serializer: &mut dyn Serializer) {
        serializer.write(match self {
            Nullable::Null(_) => &HASH_ZEROS,
            Nullable::NotNull(point) => &point.point,
        })
    }
}

impl<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> Mentionable<'a, Ctx> for Nullable<'a, Ctx, A> {
    type Fctr = NullableFactory<'a, Ctx, A>;

    fn factory(&self) -> Self::Fctr {
        match self {
            Nullable::Null(factory) => NullableFactory {
                factory: factory.clone(),
            },
            Nullable::NotNull(point) => NullableFactory {
                factory: point.origin.factory(),
            },
        }
    }

    fn points(&self) -> Vec<Point<'a, Ctx, TypelessMentionable<'a, Ctx>>> {
        match self {
            Nullable::Null(_) => vec![],
            Nullable::NotNull(point) => vec![point.typeless()],
        }
    }
}

impl<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> Clone for NullableFactory<'a, Ctx, A> {
    fn clone(&self) -> Self {
        Self {
            factory: self.factory.clone(),
        }
    }
}

impl<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> Factory<'a, Ctx>
    for NullableFactory<'a, Ctx, A>
{
    type Mtbl = Nullable<'a, Ctx, A>;

    type ParseError = PointParseError;

    fn deserialize(
        &self,
        deserializer: &mut dyn Deserializer,
        resolver: std::rc::Rc<dyn Resolver<'a, Ctx>>,
    ) -> ParseResult<'a, Ctx, Self> {
        let point = deserializer.read_n_const::<HASH_SIZE>()?;
        let factory = self.factory.clone();
        Ok(if point == HASH_ZEROS {
            Nullable::Null(factory)
        } else {
            let mut addresses = Addresses::start();
            Nullable::NotNull(addresses.next_point(deserializer, resolver, factory)?)
        })
    }

    fn unexpected_tail(&self, tail: &[u8]) -> Self::ParseError {
        PointParseError::WrongLength(HASH_SIZE + tail.len())
    }
}
